"""Global variables"""

APP_CONF = "/etc/ansible-deployer"
CFG_PERMISSIONS = "0o644"
SUBCOMMANDS = ("run", "lock", "unlock", "verify", "show")
PRINT_FAIL = '\033[91m'
PRINT_END = '\033[0m'
OPTION_EXPANSION = {
    "infra": "infrastructure",
    "raw_output": "raw_runner_output",
    "conf_val": "conf_validation"
}
ANSIBLE_DEFAULT_CALLBACK_PLUGIN_PATH = '~/.ansible/plugins/callback:/usr/share/ansible/plugins/' \
                                       'callback'



"""Module for configuration files handling"""

import os
import sys
import stat
import yaml
from cerberus import Validator
from ansible_deployer.modules.globalvars import APP_CONF, CFG_PERMISSIONS
from ansible_deployer.modules.configs.schema import SCHEMAS


class Config:
    """Class handling global configuration, and other configs: tasks, infrastructures,
    permissions"""

    def __init__(self, logger, conf_dir):
        self.logger = logger
        self.conf_dir = conf_dir if conf_dir else APP_CONF
        self.conf = self.load_global_configuration()

    def load_configuration_file(self, config_path: str):
        """Function responsible for single file loading and validation"""
        #TODO: Add verification of owner/group/persmissions
        if config_path == APP_CONF:
            self.check_cfg_permissions_and_owner(config_path)
        config_file = os.path.basename(config_path)
        self.logger.debug("Loading :%s", config_file)

        with open(config_path, "r", encoding="utf8") as config_stream:
            try:
                config = yaml.safe_load(config_stream)
            except yaml.YAMLError as e:
                self.logger.critical("Yaml loading failed for %s due to %s.", config_path, e)
                sys.exit(51)

        schema_name = config_file.split(".")[0]
        validator = Validator(SCHEMAS[schema_name])
        if not validator.validate(config, SCHEMAS[schema_name]):
            self.logger.critical("Yaml validation failed for %s due to %s.", config_path,
                                 validator.errors)
            sys.exit(53)

        self.logger.debug("Loaded:\n%s", str(config))
        return config

    def check_cfg_permissions_and_owner(self, cfg_path: str):
        """Function to verify permissions and owner for config files"""
        stat_info = os.stat(cfg_path)

        if stat_info.st_uid == 0:
            if oct(stat.S_IMODE(stat_info.st_mode)) == CFG_PERMISSIONS:
                self.logger.debug("Correct permissions and owner for config file %s.", cfg_path)
            else:
                self.logger.error("File %s permissions are incorrect! Contact your sys admin.",
                                  cfg_path)
                self.logger.error("Program will exit now.")
                sys.exit(40)
        else:
            self.logger.error("File %s owner is not root! Contact your sys admin.", cfg_path)
            self.logger.error("Program will exit now.")
            sys.exit(41)

    def get_config_paths(self):
        """Function to create absolute config paths and check their extension compatibility"""
        ymls = []
        yamls = []
        infra_cfg = None
        tasks_cfg = None
        acl_cfg = None

        for config in os.listdir(self.conf_dir):
            if config != "ansible-deploy.yaml":
                if config.endswith(".yml"):
                    ymls.append(config)
                elif config.endswith(".yaml"):
                    yamls.append(config)

                if config.startswith("infra"):
                    infra_cfg = os.path.join(self.conf_dir, config)
                elif config.startswith("tasks"):
                    tasks_cfg = os.path.join(self.conf_dir, config)
                elif config.startswith("acl"):
                    acl_cfg = os.path.join(self.conf_dir, config)

        if len(ymls) > 0 and len(yamls) > 0:
            self.logger.debug("Config files with yml extensions: %s", " ".join(ymls))
            self.logger.debug("Config files with yaml extensions: %s", " ".join(yamls))
            self.logger.critical("Config files with different extensions (.yml and .yaml) are not"
                                 " allowed in conf dir %s !", self.conf_dir)
            sys.exit(42)

        if not infra_cfg:
            self.logger.critical("Infrastructure configuration file infra.yaml does not exist in"
                                 " %s!", self.conf_dir)
            sys.exit(43)

        if not tasks_cfg:
            self.logger.critical("Tasks configuration file tasks.yaml does not exist in %s!",
                                 self.conf_dir)
            sys.exit(44)

        if not acl_cfg:
            self.logger.critical("Permission configuration file acl.yaml does not exist in %s!",
                                 self.conf_dir)
            sys.exit(45)

        return infra_cfg, tasks_cfg, acl_cfg

    def load_configuration(self):
        """Function responsible for reading configuration files and running a schema validator
        against it"""
        self.logger.debug("load_configuration called")
        #TODO: validate files/directories permissions - should be own end editable only by
        #special user
        infra_cfg, tasks_cfg, acl_cfg = self.get_config_paths()

        infra = self.load_configuration_file(infra_cfg)
        tasks = self.load_configuration_file(tasks_cfg)
        acl = self.load_configuration_file(acl_cfg)

        config = {}
        config["infra"] = infra["infrastructures"]
        config["tasks"] = tasks

        config["acl"] = {}
        for group in acl["acl_lists"]:
            key = group["name"]
            group.pop("name")
            config["acl"][key] = group

        return config

    def load_global_configuration(self):
        """Function responsible for single file loading and validation"""
        main_config_file = os.path.join(self.conf_dir, "ansible-deploy.yaml")
        if self.conf_dir == APP_CONF:
            self.check_cfg_permissions_and_owner(main_config_file)
        try:
            with open(main_config_file, "r", encoding="utf8") as config_stream:
                try:
                    config = yaml.safe_load(config_stream)
                    return config
                except yaml.YAMLError as e:
                    self.logger.critical(e, file=sys.stderr)
                    sys.exit(51)
        except FileNotFoundError:
            self.logger.critical("Main configuration file ansible-deploy.yaml does not exist in"
                                 " %s!", self.conf_dir)
            sys.exit(58)
        except Exception as exc:
            self.logger.critical("Main configuration file %s could not be opened due to: %s",
                                 main_config_file, exc)
            sys.exit(59)



"""Module for handling of logging and log files"""

import os
import logging
from logging import handlers as log_han


class Loggers:
    """Class handling creating logger and logging handlers"""

    def __init__(self, options: dict):
        self.logger = self.set_logging(options)

    @staticmethod
    def set_logging(options: dict):
        """Function to create logging objects"""
        logger = logging.getLogger("ansible-deployer_log")
        logger.setLevel(logging.DEBUG)
        formatter = logging.Formatter("%(asctime)s [%(levelname)s]: %(message)s")
        console_formatter = "\n%(asctime)s [%(levelname)s]: %(message)s\n"

        if options["syslog"]:
            rsys_handler = log_han.SysLogHandler(address="/dev/log")
            rsys_handler.setFormatter(formatter)
            rsys_handler.setLevel(logging.WARNING)
            logger.addHandler(rsys_handler)

        memory_handler = log_han.MemoryHandler(15, flushLevel=logging.DEBUG, flushOnClose=False)
        memory_handler.setFormatter(formatter)
        memory_handler.setLevel(logging.DEBUG)
        logger.addHandler(memory_handler)

        console_handler = logging.StreamHandler()
        console_handler.setFormatter(logging.Formatter(console_formatter) if options["no_color"] \
                                     else CustomFormatter(console_formatter))
        console_handler.setLevel(logging.DEBUG if options["debug"] else logging.INFO)
        logger.addHandler(console_handler)

        return logger

    def set_logging_to_file(self, log_dir: str, timestamp: str, conf: dict):
        """Function adding file handler to existing logger"""
        os.makedirs(log_dir, exist_ok=True)
        log_path = os.path.join(log_dir,
                                conf["file_naming"]["log_file_name_frmt"].format(timestamp))
        file_handler = logging.FileHandler(log_path)
        file_handler.setFormatter(logging.Formatter("%(asctime)s [%(levelname)s]: %(message)s"))
        file_handler.setLevel(logging.DEBUG)
        self.logger.addHandler(file_handler)

        return log_path

    def flush_memory_handler(self, subcommand_flag: bool, syslog: bool):
        """Flush initial log messages from memory handler to logfile"""
        if syslog:
            # handler[1] - MemoryHandler, handler[3] - FileHandler
            if subcommand_flag:
                self.logger.handlers[1].setTarget(self.logger.handlers[3])
                self.logger.handlers[1].flush()
            self.logger.handlers[1].close()
        else:
            # handler[0] - MemoryHandler, handler[2] - FileHandler
            if subcommand_flag:
                self.logger.handlers[0].setTarget(self.logger.handlers[2])
                self.logger.handlers[0].flush()
            self.logger.handlers[0].close()

class CustomFormatter(logging.Formatter):
    """Class adding colours to console logger"""

    def __init__(self, formatter):
        super().__init__()
        grey = "\x1b[0;38m"
        light_green = "\x1b[1;32m"
        yellow = "\x1b[0;33m"
        red = "\x1b[0;31m"
        light_red = "\x1b[1;31m"
        reset = "\x1b[0m"

        self.FORMATS = {
            logging.DEBUG: light_green + formatter + reset,
            logging.INFO: grey + formatter + reset,
            logging.WARNING: yellow + formatter + reset,
            logging.ERROR: red + formatter + reset,
            logging.CRITICAL: light_red + formatter + reset
        }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)



"""Module for validating input data, configuration files and data structures"""

import sys
import os
import re
from ansible_deployer.modules import globalvars


class Validators:
    """Class validating input data, yaml configuration files and their compatibility"""

    def __init__(self, logger):
        self.logger = logger

    @staticmethod
    def verify_subcommand(command: str, color_flag: bool):
        """Function to check the first arguments for a valid subcommand"""
        if color_flag:
            PRINT_FAIL = PRINT_END = ""
        else:
            PRINT_FAIL = globalvars.PRINT_FAIL
            PRINT_END = globalvars.PRINT_END

        if command not in globalvars.SUBCOMMANDS:
            print(f"{PRINT_FAIL}[CRITICAL]: Unknown subcommand :%s {PRINT_END}", (command),
                  file=sys.stderr)
            sys.exit("55")

    @staticmethod
    def verify_switches(switches: list, color_flag: bool):
        """
        Check if 2nd and following positional arguments are valid
        """
        if color_flag:
            PRINT_FAIL = PRINT_END = ""
        else:
            PRINT_FAIL = globalvars.PRINT_FAIL
            PRINT_END = globalvars.PRINT_END

        if switches[0] != "show" and len(switches[1:]) > 0:
            print(f"{PRINT_FAIL}[CRITICAL]: Too many positional arguments! Only subcommand \"show\""
                  f" can accept following arguments: all, task, infra.{PRINT_END}")
            sys.exit("56")

        for switch in switches[1:]:
            if switch not in ("all", "task", "infra"):
                print(f"{PRINT_FAIL}[CRITICAL]: Invalid argument {switch}! Subcommand \"show\" can"
                      f" accept only following arguments: all, task, infra.{PRINT_END}")
                sys.exit("57")

    def validate_options(self, options: dict):
        """Function checking if the options set are allowed in this subcommand"""
        self.logger.debug("validate_options running for subcommand: %s", options["subcommand"])
        required = []
        notsupported = []

        if options["subcommand"] == "run":
            required = ["task", "infra", "stage"]
            notsupported = ["switches"]
        elif options["subcommand"] == "verify":
            required = ["task", "infra", "stage"]
            notsupported = ["switches", "commit", "check_mode", "dry_mode"]
        elif options["subcommand"] in ("lock", "unlock"):
            required = ["infra", "stage"]
            notsupported = ["switches", "task", "commit", "keep_locked", "limit", "raw_output",
                            "self_setup", "check_mode", "dry_mode"]
        elif options["subcommand"] == "show":
            notsupported = ["task", "infra", "stage", "commit", "conf_val", "keep_locked", "syslog",
                            "limit", "raw_output", "self_setup", "check_mode", "dry_mode"]

        failed = False
        for req in required:
            if not options[req]:
                self.logger.error("Option %s is required for %s", self.expand_option_name(req),
                                  options["subcommand"])
                failed = True

        for notsup in notsupported:
            if options[notsup]:
                self.logger.error("Option %s is not supported by %s",
                                  self.expand_option_name(notsup), options["subcommand"])
                failed = True

        if failed:
            self.logger.critical("Failed to validate options")
            sys.exit(55)

    def validate_option_by_dict_with_name(self, optval: str, conf: dict):
        """
        Validate if given dictionary contains element with name equal to optval
        """
        elem = None
        if optval:
            for elem in conf:
                if elem["name"] == optval:
                    break
            else:
                self.logger.critical("%s not found in configuration file.", optval)
                sys.exit(54)

        return elem

    @staticmethod
    def validate_user_infra_stage():
        """Function checking if user has rights to execute command on selected infrastructure
        Required for: run, lock and unlock operations
        Exit on failure, return inventory on success
        """
        inventory = ""

        return inventory

    @staticmethod
    def validate_user_task():
        """Function checking if user has rights to execute the task
        Rquired for: run"""

        return True

    def validate_checkout(self, options: dict, config: dict):
        """Validate --commit or --self-setup options and return commit/path value"""
        if options["commit"] and options["self_setup"]:
            self.logger.critical("Options --commit and --self-setup are mutually exlcusive!")
            sys.exit(58)
        elif options["self_setup"]:
            commit = self.validate_self_setup(options, config)
        else:
            commit = self.validate_commit(options, config)

        return commit

    def validate_self_setup(self, options: dict, config: dict):
        """Validate if --self-setup is allowed and if requested path exists"""
        if os.path.exists(options["self_setup"]):
            for infra in config["infra"]:
                if infra["name"] == options["infra"]:
                    for stage in infra["stages"]:
                        if stage["name"] == options["stage"]:
                            allow = stage.get("allow_user_checkout", None)
                            if not allow:
                                self.logger.critical("Self setup is not allowed for infra %s!",
                                                     infra["name"])
                                sys.exit(59)
                            else:
                                return options["self_setup"]
        else:
            self.logger.critical("Provided --self-setup path %s does not exist!",
                                 options["self_setup"])
            sys.exit(59)
        return None

    def validate_commit(self, options: dict, config: dict):
        """Function to validate commit value against config and assign final value"""
        if not options["commit"]:
            commit_id = None
            self.logger.debug("Using default commit.")
        else:
            for item in config["tasks"]["tasks"]:
                if item["name"] == options["task"]:
                    for elem in item["allowed_for"]:
                        available_commits = elem.get("commit", [])
                        for commit in available_commits:
                            commit_id = re.fullmatch(commit, options["commit"])
                            if commit_id:
                                commit_id = commit_id.group()
                                self.logger.debug("Using commit: %s .", commit_id)
                                break
                        else:
                            continue
                        break
                    else:
                        self.logger.error("Requested commit %s is not valid for task %s.",
                                     options["commit"], options["task"])
                        self.logger.error("Program will exit now.")
                        sys.exit(56)

        return commit_id

    def validate_option_values_against_config(self, config: dict, options: dict):
        """
        Function responsible for checking if option values match configuration
        """
        selected_items = {}
        for option in options.keys():
            if options[option]:
                if option == "infra":
                    selected_items["infra"] = self.validate_option_by_dict_with_name(
                                              options["infra"], config["infra"])
                elif option == "task":
                    selected_items["task"] = self.validate_option_by_dict_with_name(options["task"],
                                             config["tasks"]["tasks"])
                elif option == "stage":
                    for item in config["infra"]:
                        if item["name"] == options["infra"]:
                            index = config["infra"].index(item)
                    selected_items["stage"] = self.validate_option_by_dict_with_name(
                                              options["stage"], config["infra"][index]["stages"])
                elif option == "limit":
                    for item in config["tasks"]["tasks"]:
                        if item["name"] == options["task"]:
                            if self.get_limit_flag(options, item):
                                selected_items["limit"] = options["limit"]
                                break
                    else:
                        self.logger.critical("Limit %s is not available for task %s.",
                                             options["limit"], options["task"])
                        sys.exit(54)

        selected_items["commit"] = self.validate_checkout(options, config)
        self.logger.debug("Completed validate_option_values_with_config")

                #TODO: validate if user is allowed to use --commit
                #TODO: validate if user is allowed to execute the task on infra/stag pair
                #(validate_user_infra_stage(), validate_usr_task())

        return selected_items

    def verify_task_permissions(self, selected_items: dict, user_groups: list, config: dict):
        """
        Function verifies if the running user is allowed to run the task
        """
        s_task = selected_items["task"]
        s_infra = selected_items["infra"]
        o_stage = selected_items["stage"]
        self.logger.debug("Running verify_task_permissions for s_task:\n%s,\ns_infra:\n%s,\n"
                          "o_stage:\n%s\n and user groups:\n%s", s_task, s_infra, o_stage,
                          user_groups)

        for item in s_task["allowed_for"]:
            acl_group = item["acl_group"]
            self.logger.debug("\tChecking permission group: %s, for user_groups: %s", acl_group,
                              user_groups)
            self.logger.debug("\tPermission group %s content: %s", acl_group,
                              str(config["acl"][acl_group]))
            if any(group in user_groups for group in config["acl"][acl_group]["groups"]):
                for infra in config["acl"][acl_group]["infra"]:
                    self.logger.debug("\t\tChecking infra: %s for infra: %s", infra,
                                 selected_items["infra"]["name"])
                    if infra["name"] == selected_items["infra"]["name"]:
                        for stage in infra["stages"]:
                            self.logger.debug("\t\t\tChecking stage: %s for stage: %s", stage,
                                              o_stage["name"])
                            if stage == o_stage["name"]:
                                self.logger.debug("Task allowed, based on %s", acl_group)
                                return True
        self.logger.debug("Task forbidden")
        return False

    @staticmethod
    def expand_option_name(option: str):
        """Expand name of option variable name to option argument name"""
        if option in ["conf_val", "infra", "raw_output"]:
            option = globalvars.OPTION_EXPANSION[option]

        formatted_option = "--" + option.replace("_", "-")
        return formatted_option

    @staticmethod
    def get_limit_flag(options: dict, task: dict):
        """Evaluate if runner option --limit should be allowed for this task based on cli options"""
        if options["check_mode"]:
            return True
        return task.get("allow_limit", False)



"""Module for formatting outputs from different sources"""

class Formatters:
    """Class handling formatting of ansible standard output and error streams"""

    error_indicators = ["error", "fatal"]
    task_indicators = ["task", "running handler"]

    def __init__(self, logger):
        self.logger = logger

    def positive_ansible_output(self, warning: list, output: list, command: str):
        """Log output for a positive case in ansible execution"""
        if warning:
            self.logger.warning("\n".join(warning))
        if output:
            self.logger.info("\n".join(output))
        self.logger.info("\"%s\" ran succesfully", " ".join(command))

    def negative_ansible_output(self, warning: list, error: list, command: str):
        """Log output for a negative case in ansible execution"""
        if warning:
            self.logger.warning("\n".join(warning))
        self.logger.error("\"%s\" failed due to:", " ".join(command))
        if error:
            self.logger.error("\n".join(error))

    def format_std_out(self, std_out):
        """Decode standard output to logger.info"""
        for line in std_out.split(b"\n\n"):
            self.logger.info(line.decode("utf-8"))

    def format_std_err(self, std_err):
        """Decode standard error to logger.error"""
        for line in std_err.split(b"\n\n"):
            self.logger.error(line.decode("utf-8"))

    def debug_std_out(self, std_out):
        """Decode standard output to logger.debug"""
        for line in std_out.split(b"\n\n"):
            self.logger.debug(line.decode("utf-8"))

    def debug_std_err(self, std_err):
        """Decode standard error to logger.debug"""
        for line in std_err.split(b"\n\n"):
            self.logger.debug(line.decode("utf-8"))

    def format_ansible_output(self, process_output: list):
        """Group and format output from ansible execution"""
        std_output = []
        std_warning = []
        std_error = []
        std_complete = []

        for no, line in enumerate(process_output):
            if any(eindicator in line.lower() for eindicator in self.error_indicators):
                for pline in process_output[no-2:no]:
                    if any(tindicator in pline.lower() for tindicator in self.task_indicators)\
                            or pline == "":
                        std_error.append(pline)
                std_error.append(line)
                for nline in process_output[no+1:self.find_end_of_task(process_output[no+1:],
                                                                       no+1)]:
                    std_error.append(nline)
                std_complete.append(line)
            elif "warn" in line.lower():
                std_warning.append(line)
            else:
                std_output.append(line)
                std_complete.append(line)

        return {
            "output": std_output,
            "warning": std_warning,
            "error": std_error,
            "complete": std_complete
        }

    @staticmethod
    def find_end_of_task(stream_fragment: list, parent_index: int):
        """Parse list of output elements to find beginning of next task and end of current task."""
        for no, line in enumerate(stream_fragment):
            if "changed=true" in line or "changed=false" in line or not line:
                return parent_index + no
        return None



"""Module for creating, initialization and connecting to database"""
import os
import sys
import sqlite3
import logging


class DbSetup:
    """Class handling creation and initialization of database"""

    def __init__(self, logger: logging, start_ts: str, conf: dict, options: dict):
        self.logger = logger
        self.start_ts = start_ts
        self.conf = conf
        self.options = options
        self.db_path = self.create_db_path()

    def create_db_path(self):
        """Create absolute path for database"""
        short_ts = self.start_ts.split("_")[0]
        db_name = self.options["infra"] + "_" + self.options["stage"] + ".db"
        return os.path.join(os.path.join(self.conf["global_paths"]["work_dir"], short_ts,
                                         self.conf["global_paths"]["database_subdir"], db_name))

    def create_db_dir(self):
        """Create directory containing databases"""
        db_dir = os.path.dirname(self.db_path)
        try:
            os.mkdir(db_dir)
            try:
                os.chmod(db_dir, int(self.conf["permissions"]["parent_workdir"], 8))
                self.logger.debug("Successfully created database directory: %s .", db_dir)
            except Exception as exc:
                self.logger.critical("Failed to change permissions of database directory: %s error"
                                     " was: %s", db_dir, exc)
                sys.exit(100)
        except FileExistsError:
            self.logger.debug("Database directory already exists: %s .", db_dir)
        except Exception as exc:
            self.logger.critical("Failed to create database directory: %s error was: %s.", db_dir,
                                 exc)
            sys.exit(100)

    def connect_to_db(self):
        """Create connector to sqlite database"""
        self.create_db_dir()
        try:
            connection = sqlite3.connect(self.db_path)
            self.logger.debug("Connection opened to database: %s .", self.db_path)
            db_perm_string = f"0o{''.join(str(oct(os.stat(self.db_path).st_mode))[-4:])}"
            if db_perm_string != self.conf["permissions"]["parent_workdir"]:
                self.change_db_permissions()
        except Exception as exc:
            self.logger.critical("Failed to connect to database: %s error was: %s.", self.db_path,
                                 exc)
            sys.exit(101)

        return connection, self.db_path

    def change_db_permissions(self):
        """Ensure correct permissions are set when database file is created"""
        try:
            os.chmod(self.db_path, int(self.conf["permissions"]["parent_workdir"], 8))
            self.logger.debug("Database %s permissions changed.", self.db_path)
        except Exception as exc:
            self.logger.critical("Failed to change permissions of database: %s error was: %s",
                                 self.db_path, exc)
            sys.exit(101)



"""Sqlite database schema dictionaries"""

SCHEMAS = {
    "sequences": {
        "sequence_id": "text",
        "hostname": "text",
        "user_id": "text",
        "hooks": "text",
        "task": "text",
        "start": "text",
        "end": "text",
        "host_locked": "text",
        "keep_locked_on": "text",
        "self_setup_on": "text",
        "conf_dir_on": "text"
    },
    "play_item_tasks": {
        "task_id": "integer primary key autoincrement",
        "sequence_id": "text",
        "task_name": "text",
        "result": "text",
        "hostname": "text",
        "timestamp": "text",
        "task_details": "text"
    }
}



"""Ansible callback plugin that logs to a file, based on community.general.log_plays 2.0"""
# -*- coding: utf-8 -*-

# pylint: disable=duplicate-code

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

import os
import time
import json

from ansible.utils.path import makedirs_safe
from ansible.module_utils.common.text.converters import to_bytes
from ansible.module_utils.common._collections_compat import MutableMapping
from ansible.parsing.ajson import AnsibleJSONEncoder
from ansible.plugins.callback import CallbackBase


class CallbackModule(CallbackBase):
    """
    logs playbook results in defined logfile
    """
    CALLBACK_VERSION = 2.0
    CALLBACK_TYPE = 'notification'
    CALLBACK_NAME = 'log_plays_adjusted'
    CALLBACK_NEEDS_WHITELIST = True

    TIME_FORMAT = "%b %d %Y %H:%M:%S"
    MSG_FORMAT = "%(now)s - %(playbook)s - %(task_name)s - %(category)s\n%(data)s\n\n"

    def __init__(self):

        super().__init__()
        self.log_path = os.environ["LOG_PLAYS_PATH"]

    def set_options(self, task_keys=None, var_options=None, direct=None):
        """Set ansible options and try to create log directory"""
        super().set_options(task_keys=task_keys, var_options=var_options, direct=direct)

        log_dir = os.path.dirname(self.log_path)
        if not os.path.exists(log_dir):
            makedirs_safe(log_dir)

    def log(self, result, category):
        """Log message to log file"""
        data = result._result
        if isinstance(data, MutableMapping):
            if '_ansible_verbose_override' in data:
                # avoid logging extraneous data
                data = 'omitted'
            else:
                data = data.copy()
                invocation = data.pop('invocation', None)
                data = json.dumps(data, cls=AnsibleJSONEncoder)
                if invocation is not None:
                    data = f"{json.dumps(invocation)} => {data} "

        now = time.strftime(self.TIME_FORMAT, time.localtime())

        msg = to_bytes(
            self.MSG_FORMAT
            % {
                "now": now,
                "playbook": self.playbook,
                "task_name": result._task.name,
                "category": category,
                "data": data
            }
        )
        with open(self.log_path, "ab") as fd:
            fd.write(msg)

    def v2_runner_on_failed(self, result, ignore_errors=False):
        """Log result when runner failed"""
        self.log(result, 'FAILED')

    def v2_runner_on_ok(self, result):
        """Log result when runner passed"""
        self.log(result, 'OK')

    def v2_runner_on_skipped(self, result):
        """Log result when runner skipped host"""
        self.log(result, 'SKIPPED')

    def v2_runner_on_unreachable(self, result):
        """Log result when host was unreachable"""
        self.log(result, 'UNREACHABLE')

    def v2_runner_on_async_failed(self, result):
        """Log result when async failed"""
        self.log(result, 'ASYNC_FAILED')

    def v2_playbook_on_start(self, playbook):
        """Log runner start"""
        # pylint: disable=attribute-defined-outside-init
        self.playbook = playbook._file_name

    def v2_playbook_on_import_for_host(self, result, imported_file):
        """Log runner import"""
        self.log(result, 'IMPORTED')

    def v2_playbook_on_not_import_for_host(self, result, missing_file):
        """Log runner not import"""
        self.log(result, 'NOTIMPORTED')



"""Module for writing data to existing database"""
from itertools import chain
import json
import sys
import os
import datetime
import sqlite3
from ansible_deployer.modules.database.schema import SCHEMAS


class DbWriter:
    """Class handling writing records to database"""

    def __init__(self, logger, connector, db_path: str):
        self.logger = logger
        self.connector = connector
        self.db_path = db_path
        self.cursor = self.connector.cursor()
        self.create_tables()

    def create_tables(self):
        """Create database tables for all keys in schema"""
        for table, columns in SCHEMAS.items():
            self.create_table(table, columns)

    def create_table(self, table_name: str, columns: dict):
        """Create single database table"""
        base_format = str(len(columns) * "'{}' '{}', ").strip(", ")
        table_string = f'''CREATE TABLE {table_name} ({base_format})'''
        try:
            self.cursor.execute(table_string.format(*list(chain(*columns.items()))))
            self.logger.debug("Created table %s in database %s .", table_name, self.db_path)
        except sqlite3.OperationalError:
            self.logger.debug("Table %s already exists in database %s .", table_name, self.db_path)
        except Exception as exc:
            self.logger.critical("Failed table %s creation in database %s, error was %s",
                                 table_name, self.db_path, exc)
            sys.exit(104)

    def parse_yaml_output_for_hosts(self, stream: list, sequence_id: str):
        """Get all hosts used in runner from runner output"""
        # TODO should be taken from better source
        record_hosts = []
        try:
            for line in stream:
                if "changed=true" in line or "changed=false" in line:
                    record_hosts.append(line.split("[")[1].split("]")[0])
            return list(set(record_hosts))
        except Exception as exc:
            self.logger.critical("Failed parsing output stream to yaml, error was %s", exc)
            sys.exit(102)

    @staticmethod
    def find_end_of_task(stream_fragment: list, parent_index: int):
        """Parse list of output elements to find beginning of next task and end of current task."""
        for no, line in enumerate(stream_fragment):
            if "changed=true" in line or "changed=false" in line or not line:
                return parent_index + no
        return None

    def write_records(self, record_dict: dict):
        """Write multiple records to "play_item_tasks" table"""
        host_list = []
        for hosts in record_dict.values():
            for host, records in hosts.items():
                host_list.append(host)
                self.write_record("play_item_tasks", self.record_with_json([*records.values()]))

        return host_list

    @staticmethod
    def record_with_json(record: list):
        """Replace dict with json in record list"""
        new_record = record[:-1]
        new_record.append(json.dumps(record[-1]))
        return new_record

    def write_record(self, table_name, record):
        """Write single record to any table"""
        base_format = str(len(record) * "?, ").strip(", ")
        insert_string = f"INSERT INTO {table_name} VALUES ({base_format})"
        try:
            self.cursor.execute(insert_string, record)
        except Exception as exc:
            self.logger.critical("Failed writing record to database %s, error was %s",
                                 self.db_path, exc)
            sys.exit(105)

    def commit_changes(self):
        """Commit changes to sqlite database"""
        try:
            self.connector.commit()
            self.logger.debug("Changes saved to database %s .", self.db_path)
        except Exception as exc:
            self.logger.critical("Failed saving changes to database %s, error was %s",
                                 self.db_path, exc)
            sys.exit(103)

    def start_sequence_dict(self, hosts: list, setup_hooks: list, options: dict, start_ts_raw: str,
                            sequence_id: str):
        """Create dictionary with constant values for "sequences" table records"""
        sequence_dict = {}
        hosts = list(set(hosts))
        for host in hosts:
            sequence_dict[host] = self.fill_sequence_dict(sequence_id, host, setup_hooks, options,
                                                          start_ts_raw)

        return sequence_dict

    @staticmethod
    def fill_sequence_dict(sequence_id: str, host: str, setup_hooks: list, options: dict,
                           start_ts_raw: str):
        """Assign values to keys (columns) for sequence record dictionary"""
        sub_sequence_dict = dict(SCHEMAS["sequences"])
        sub_sequence_dict["sequence_id"] = sequence_id
        sub_sequence_dict["hostname"] = host
        sub_sequence_dict["user_id"] = os.getuid()
        hook_names = []
        for hook in setup_hooks:
            hook_names.append(hook["name"])
        sub_sequence_dict["hooks"] = ", ".join(hook_names)
        sub_sequence_dict["task"] = options["task"]
        sub_sequence_dict["start"] = start_ts_raw
        sub_sequence_dict["keep_locked_on"] = str(bool(options["keep_locked"]))
        sub_sequence_dict["self_setup_on"] = str(bool(options["self_setup"]))
        sub_sequence_dict["conf_dir_on"] = str(bool(options["conf_dir"]))

        return sub_sequence_dict

    def finalize_db_write(self, sequence_dict: dict, host_locked: bool):
        """Add final values to sequence record dictionary, commit changes and close database
        connection"""
        end_ts = datetime.datetime.now()
        for params in sequence_dict.values():
            record_list = []
            params["end"] = end_ts
            params["host_locked"] = str(bool(host_locked))
            for params2 in params.values():
                record_list.append(params2)
            self.write_record("sequences", record_list)
        self.commit_changes()
        self.connector.close()



"""
Ansible callback plugin that logs to existing play_item_tasks table in initiated sqlite database.
Created and tailored for ansible-deployer project.
"""
# -*- coding: utf-8 -*-
# Copyright (c) 2012, Michael DeHaan, <michael.dehaan@gmail.com>
# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# pylint: disable=duplicate-code

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

import json
import os
import sqlite3
import time

from ansible.module_utils.common._collections_compat import MutableMapping
from ansible.parsing.ajson import AnsibleJSONEncoder
from ansible.plugins.callback import CallbackBase


# NOTE: in Ansible 1.2 or later general logging is available without
# this plugin, just set ANSIBLE_LOG_PATH as an environment variable
# or log_path in the DEFAULTS section of your ansible configuration
# file.  This callback is an example of per hosts logging for those
# that want it.


class CallbackModule(CallbackBase):
    """
    logs playbook results to sqlite db
    """
    CALLBACK_VERSION = 1.0
    CALLBACK_TYPE = 'notification'
    CALLBACK_NAME = 'sqlite_deployer'
    CALLBACK_NEEDS_WHITELIST = True

    TABLE_NAME = "play_item_tasks"
    TABLE_COLUMNS = ["task_id", "sequence_id", "task_name", "result", "hostname", "timestamp",
                     "task_details"]
    TIME_FORMAT = "%b %d %Y %H:%M:%S"

    def __init__(self):

        super().__init__()
        self.db_dir = os.environ["SQLITE_PATH"]
        self.sequence = os.environ["SEQUENCE_ID"]
        self.connector = self.db_connect()
        self.cursor = self.connector.cursor()

    def set_options(self, task_keys=None, var_options=None, direct=None):
        super().set_options(task_keys=task_keys, var_options=var_options, direct=direct)

    def db_connect(self):
        """Create sqlite db connector"""
        return sqlite3.connect(self.db_dir)

    def log(self, result, category):
        """Log message to sqlite database"""
        data = result._result
        if isinstance(data, MutableMapping):
            if '_ansible_verbose_override' in data:
                # avoid logging extraneous data
                data = 'omitted'
            else:
                data = data.copy()
                invocation = data.pop('invocation', None)
                data = json.dumps(data, cls=AnsibleJSONEncoder)
                if invocation is not None:
                    data = json.dumps(invocation, cls=AnsibleJSONEncoder)

        now = time.strftime(self.TIME_FORMAT, time.localtime())

        try:
            self.cursor.execute(
                "INSERT INTO {} ({}, {}, {}, {}, {}, {}) VALUES(?,?,?,?,?,?);".format(
                self.TABLE_NAME, *self.TABLE_COLUMNS), (
                    self.sequence,
                    result._task.name,
                    category,
                    result._host.get_name(),
                    now,
                    data
                )
            )
            self.connector.commit()
        except Exception as exc:
            print(exc)

    def v2_runner_on_failed(self, result, ignore_errors=False):
        """Log result when runner failed"""
        self.log(result, 'FAILED')

    def v2_runner_on_ok(self, result):
        """Log result when runner passed"""
        self.log(result, 'OK')

    def v2_runner_on_skipped(self, result):
        """Log result when runner skipped host"""
        self.log(result, 'SKIPPED')

    def v2_runner_on_unreachable(self, result):
        """Log result when host was unreachable"""
        self.log(result, 'UNREACHABLE')

    def v2_runner_on_async_failed(self, result):
        """Log result when async failed"""
        self.log(result, 'ASYNC_FAILED')

    def v2_playbook_on_start(self, playbook):
        """Log runner start"""
        # pylint: disable=attribute-defined-outside-init
        self.playbook = playbook._file_name

    def v2_playbook_on_import_for_host(self, result, imported_file):
        """Log runner import"""
        self.log(result, 'IMPORTED')

    def v2_playbook_on_not_import_for_host(self, result, missing_file):
        """Log runner not import"""
        self.log(result, 'NOTIMPORTED')



"""Yaml schemas for validation"""

SCHEMAS = {
    "acl": {
            "acl_lists": {
                "type": "list",
                "schema": {
                    "type": "dict",
                    "schema": {
                        "name": {
                            "type": "string",
                            "required": True
                        },
                        "groups": {
                            "type": "list",
                            "schema": {
                                "type": "string",
                                "required": True
                            }
                        },
                        "infra": {
                            "type": "list",
                            "schema": {
                                "type": "dict",
                                "schema": {
                                    "name": {
                                        "type": "string",
                                        "required": True
                                    },
                                    "stages": {
                                        "type": "list",
                                        "required": True
                                    }
                                }
                            }
                        }
                    }
                }
            }
        },

    "infra": {
            "infrastructures": {
                "type": "list",
                "required": True,
                "schema": {
                    "type": "dict",
                    "schema": {
                        "name": {
                            "type": "string",
                            "required": True
                        },
                        "stages": {
                            "type": "list",
                            "required": True,
                            "schema": {
                                "type": "dict",
                                "schema": {
                                    "name": {
                                        "type": "string",
                                        "required": True
                                    },
                                    "inventory": {
                                        "type": "string",
                                        "required": True
                                    },
                                    "allow_user_checkout": {
                                        "type": "boolean",
                                        "required": False
                                    }
                                }
                            }
                        }
                    }
                }
            }
        },

    "tasks": {
            "setup_hooks": {
                "type": "list",
                "schema": {
                    "type": "dict",
                    "schema": {
                        "name": {
                            "type": "string",
                            "required": True
                        },
                        "module": {
                            "type": "string",
                            "required": True
                        },
                        "opts": {
                            "type": "dict",
                            "required": False
                        }
                    }
                }
            },
            "play_items": {
                "type": "list",
                "required": True,
                "schema": {
                    "type": "dict",
                    "schema": {
                        "name": {
                            "type": "string",
                            "required": True
                        },
                        "file": {
                            "type": "string",
                            "required": True
                        },
                        "runner": {
                            "type": "string",
                            "required": False,
                            "allowed": [
                                "ansible-playbook",
                                "py.test"
                            ]
                        },
                        "skip": {
                            "type": "list",
                            "required": False,
                            "schema": {
                                "type": "dict",
                                "schema": {
                                    "infra": {
                                        "type": "string",
                                        "required": True
                                    },
                                    "stage": {
                                        "type": "string",
                                        "required": True
                                    }
                                }
                            }
                        }
                    }
                }
            },
            "tasks": {
                "type": "list",
                "required": True,
                "schema": {
                    "type": "dict",
                    "required": True,
                    "schema": {
                        "name": {
                            "type": "string",
                            "required": True
                        },
                        "play_items": {
                            "type": "list",
                            "required": True,
                            "schema": {
                                "type": "string",
                                "required": True
                            }
                        },
                        "verify_items": {
                            "type": "list",
                            "required": False,
                            "schema": {
                                "type": "string",
                                "required": True
                            }
                        },
                        "allowed_for": {
                            "type": "list",
                            "required": True,
                            "schema": {
                                "type": "dict",
                                "required": True,
                                "schema": {
                                    "acl_group": {
                                        "type": "string",
                                        "required": True
                                    },
                                    "commit": {
                                        "type": "list",
                                        "required": False,
                                        "schema": {
                                            "type": "string"
                                        }
                                    }
                                }
                            }
                        },
                        "allow_limit": {
                            "type": "boolean",
                            "required": False
                        },
                        "tags": {
                            "type": "list",
                            "required": False
                        },
                        "skip_tags": {
                                "type": "list",
                                "required": False
                        }
                    }
                }
            }
        }
}



"""Module for handling infrastructure locks"""

import os
import sys
import pwd


class Locking:
    """Class handling infrastructure locks"""

    def __init__(self, logger, keep_lock, infra, conf):
        self.logger = logger
        self.keep_lock = keep_lock
        self.infra = infra
        self.conf = conf

    def lock_inventory(self, lockpath: str):
        """
        Function responsible for locking inventory file.
        The goal is to prevent two parallel ansible-deploy's running on the same inventory
        This needs to be done by the use of additional directory under PARNT_WORKDIR,, for instance:
        PARENT_WORKDIR/locks.
        We shouldn't check if file exists, but rather attempt to open it for writing, until we're
        done every other process should be rejected this access.
        The file should match inventory file name.
        """

        lockdir = os.path.dirname(lockpath)
        self.logger.debug("Started lock_inventory for lockdir: %s and lockpath %s.", lockdir,
                          lockpath)
        os.makedirs(lockdir, exist_ok=True)

        try:
            with open(lockpath, "x", encoding="utf8") as fh:
                fh.write(str(os.getpid()))
                fh.write(str("\n"))
                fh.write(str(pwd.getpwuid(os.getuid()).pw_name))
            os.chmod(lockpath, int(self.conf["permissions"]["lock"], 8))
            self.logger.info("Infra locked.")
        except FileExistsError:
            with open(lockpath, "r", encoding="utf8") as fh:
                proc_pid, proc_user = fh.readlines()
            self.logger.critical("Another process (PID: %s) started by user %s is using this"
                                " infrastructure, please try again later.", proc_pid.strip(),
                                proc_user.strip())
            sys.exit(61)
        except Exception as exc:
            self.logger.critical(exc)
            sys.exit(62)

    def unlock_inventory(self, lockpath: str):
        """
        Function responsible for unlocking inventory file, See also lock_inventory
        """

        self.logger.debug("Started unlock_inventory for lockpath %s.", lockpath)

        if not self.keep_lock:
            try:
                os.remove(lockpath)
                self.logger.info("Lock %s has been removed.", lockpath)
            except FileNotFoundError:
                self.logger.critical("Requested lock %s was not found. Nothing to do.", lockpath)
                sys.exit(63)
            except Exception as exc:
                self.logger.critical(exc)
                sys.exit(64)
        else:
            self.logger.debug("Keep locked infra %s:%s .", self.infra[0], self.infra[1])



"""Module for running system processes"""

import os
import sys
import subprocess
from ansible_deployer.modules.globalvars import ANSIBLE_DEFAULT_CALLBACK_PLUGIN_PATH
from ansible_deployer.modules.outputs.formatting import Formatters

class Runners:
    """Class handling ansible hooks and ansible plays execution"""

    def __init__(self, logger, lock_obj, workdir, start_ts_raw, setup_hooks, log_path, db_path):
        self.logger = logger
        self.lock_obj = lock_obj
        self.workdir = workdir
        self.start_ts_raw = start_ts_raw
        self.setup_hooks = setup_hooks
        self.sequence_id = os.path.basename(self.workdir)
        self.log_path = log_path
        self.db_path = db_path

    @staticmethod
    def reassign_commit_and_workdir(commit: str, workdir: str):
        """Change workdir if commit is a path (option --self-setup enabled in main)"""
        if not commit:
            commit = ""
        elif os.path.exists(os.path.abspath(commit)):
            workdir = commit
            commit = ""

        return workdir, commit

    def setup_ansible(self, commit: str, conf_dir: str):
        """
        Function responsible for execution of setup_hooks
        It passes the "commit" to the hook if one given, if not the hook should
        checkout the default repo.
        """
        failed = False
        workdir, commit = Runners.reassign_commit_and_workdir(commit, self.workdir)

        for hook in self.setup_hooks:
            if hook["module"] == "script":
                try:
                    with subprocess.Popen([os.path.join(conf_dir, hook["opts"]["file"]), commit],
                                          stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                          stdin=subprocess.PIPE, cwd=workdir) as proc:
                        std_out, std_err = proc.communicate()
                        if proc.returncode != 0:
                            failed = True
                except Exception as e:
                    self.logger.critical("Failed executing %s: %s", hook["opts"]["file"], e)
                    sys.exit(41)
                else:
                    if std_out:
                        self.logger.info("Setup hook %s stdout:", hook["name"])
                        for line in std_out.split(b"\n"):
                            if line:
                                self.logger.info(line.decode("utf-8"))
                    if std_err:
                        self.logger.error("Setup hook %s stderr:", hook["name"])
                        for line in std_err.split(b"\n"):
                            if line:
                                self.logger.error(line.decode("utf-8"))
                    if proc.returncode:
                        self.logger.critical("Setup hook %s failed, cannot continue", hook["name"])
                        sys.exit(40)
                    else:
                        self.logger.info("Setup completed in %s", workdir)

            else:
                self.logger.error("Not supported")
            if failed:
                self.logger.critical("Program will exit now.")
                sys.exit(69)

    def get_playitems(self, config: dict, options: dict):
        """
        Function obtaining play items for specified task.
        :param config:
        :param task_name:
        :return:
        """
        playitems = []

        for item in config["tasks"]["tasks"]:
            if item["name"] == options["task"]:
                if options["subcommand"] == "run":
                    play_names = item["play_items"]
                elif options["subcommand"] == "verify":
                    play_names = item["verify_items"]
                else:
                    self.logger.critical("Should have never happen. Uncleaned lock left")
                    sys.exit(130)

        for play in play_names:
            for item in config["tasks"]["play_items"]:
                if item["name"] == play:
                    skip = item.get("skip", [])
                    if skip:
                        for elem in item["skip"]:
                            if elem["infra"] == options["infra"] and \
                                elem["stage"] == options["stage"]:
                                self.logger.info("Skipping playitem %s on %s and %s stage.", play,
                                            options["infra"], options["stage"])
                                break
                        else:
                            playitems.append(item)
                    else:
                        playitems.append(item)

        # TODO add check if everything was skipped
        return playitems

    def run_playitem(self, config: dict, options: dict, inventory: str, lockpath: str, db_writer):
        """
        Function implementing actual execution of runner [ansible-playbook or py.test]
        """
        playitems = self.get_playitems(config, options)
        host_list = []
        returned = []
        if not playitems:
            self.logger.critical("No playitems found for requested task %s. Nothing to do.",
                                 options['task'])
            self.lock_obj.unlock_inventory(lockpath)
            sequence_records = db_writer.start_sequence_dict([""], self.setup_hooks, options,
                                                             self.start_ts_raw, self.sequence_id)
            db_writer.finalize_db_write(sequence_records, False)
            sys.exit(70)
        else:
            for playitem in playitems:
                command, command_env = self.construct_command(playitem, inventory, config, options)
                self.logger.debug("Running '%s'.", command)
                try:
                    with subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                                          env=command_env) as proc:

                        for msg in proc.stdout:
                            dec_msg = msg.split(b"\n")[0].decode("utf-8")
                            returned.append(dec_msg)
                            if options["raw_output"]:
                                print(dec_msg)

                        proc.communicate()
                        format_obj = Formatters(self.logger)
                        parsed_std = format_obj.format_ansible_output(returned)
                        host_list = db_writer.parse_yaml_output_for_hosts(parsed_std["complete"],
                                                                self.sequence_id)
                        sequence_records = db_writer.start_sequence_dict(host_list,
                                                                         self.setup_hooks, options,
                                                                         self.start_ts_raw,
                                                                         self.sequence_id)

                        if proc.returncode == 0:
                            format_obj.positive_ansible_output(parsed_std["warning"],
                                                               parsed_std["output"], command)
                        else:
                            format_obj.negative_ansible_output(parsed_std["warning"],
                                                               parsed_std["error"], command)
                            self.lock_obj.unlock_inventory(lockpath)
                            db_writer.finalize_db_write(sequence_records, False)
                            self.logger.critical("Program will exit now.")
                            sys.exit(71)
                except Exception as exc:
                    self.logger.critical("\"%s\" failed due to:")
                    self.logger.critical(exc)
                    sequence_records = db_writer.start_sequence_dict(host_list, self.setup_hooks,
                                                                     options, self.start_ts_raw,
                                                                     self.sequence_id)
                    db_writer.finalize_db_write(sequence_records, True)
                    sys.exit(72)
            return sequence_records

    @staticmethod
    def get_tags_for_task(config: dict, options: dict):
        """Function to get task's tags"""
        tags = []
        skip_tags = []
        for task in config["tasks"]["tasks"]:
            if task["name"] == options["task"]:
                tags = task.get("tags", [])

        if options["dry_mode"]:
            tags.append("ansible_deployer_dry_mode")

        for task in config["tasks"]["tasks"]:
            if task["name"] == options["task"]:
                skip_tags = task.get("skip_tags", [])

        return tags, skip_tags

    def construct_command(self, playitem: str, inventory: str, config: dict, options: dict):
        """Create final ansible command from available variables"""
        tags, skip_tags = Runners.get_tags_for_task(config, options)

        if "runner" in playitem and playitem["runner"] == "py.test":
            command = ["py.test", "--ansible-inventory", inventory]
            if options["limit"]:
                command.append(f"--hosts=ansible://{options['limit']}")
            else:
                command.append("--hosts=ansible://all")
            command.append("--junit-xml=junit_"+options['task']+'.xml')
            command.append("./"+playitem["file"])
            command_env = os.environ
        else:
            command = ["ansible-playbook", "-v", "-i", inventory, playitem["file"]]
            if options["limit"]:
                command.append("-l")
                command.append(options["limit"])
            if tags:
                command.append("-t")
                command.append(",".join(tags))
            if skip_tags:
                command.append("--skip-tags")
                command.append(",".join(skip_tags))
            if options["check_mode"]:
                command.append("-C")
            command_env=dict(os.environ, ANSIBLE_STDOUT_CALLBACK="yaml", ANSIBLE_NOCOWS="1",
                             ANSIBLE_LOAD_CALLBACK_PLUGINS="1", LOG_PLAYS_PATH=self.log_path,
                             ANSIBLE_CALLBACKS_ENABLED="log_plays_adjusted,sqlite_deployer",
                             ANSIBLE_CALLBACK_PLUGINS=self.append_to_ansible_callbacks_path(),
                             SQLITE_PATH=self.db_path, SEQUENCE_ID=self.sequence_id)

        return command, command_env

    @staticmethod
    def append_to_ansible_callbacks_path():
        """Create final searchable path for ansible callback plugins"""
        plugin_path = os.path.join(os.path.realpath(__file__).rsplit(os.sep, 3)[0], "plugins")
        return f'{ANSIBLE_DEFAULT_CALLBACK_PLUGIN_PATH}:{plugin_path}'



"""Main module for ansible-deploy"""

import os
import sys
import argparse
import datetime
import errno
import pkg_resources
from ansible_deployer.modules.configs.config import Config
from ansible_deployer.modules.locking.locking import Locking
from ansible_deployer.modules.outputs.logging import Loggers
from ansible_deployer.modules.validators.validate import Validators
from ansible_deployer.modules.runners.run import Runners
from ansible_deployer.modules.database.creator import DbSetup
from ansible_deployer.modules.database.writer import DbWriter
from ansible_deployer.modules import misc
from ansible_deployer.modules import globalvars

def parse_options(argv):
    """Generic function to parse options for all commands, we validate if the option was allowed for
    specific subcommand outside"""
    parser = argparse.ArgumentParser(add_help=True)

    parser.add_argument("subcommand", nargs='*', default=None, metavar="SUBCOMMAND",
                        help='Specify subcommand to execute. Available commands: '+ \
                        str(globalvars.SUBCOMMANDS))
    parser.add_argument("--infrastructure", "-i", nargs=1, default=[None], metavar="INFRASTRUCTURE",
                        help='Specify infrastructure for deploy.')
    parser.add_argument("--inventory", "-I", nargs=1, default=[None], metavar="INVENTORY",
                        help='Specify inventory, only allowed if not set in configuration')
    parser.add_argument("--stage", "-s", nargs=1, default=[None], metavar="STAGE",
                        help='Specify stage type. Available types are: "testing" and "prod".')
    parser.add_argument("--commit", "-c", nargs=1, default=[None], metavar="COMMIT",
                        help='Provide commit ID.')
    parser.add_argument("--task", "-t", nargs=1, default=[None], metavar='TASK_NAME',
                        help='Provide task_name.')
    parser.add_argument("--conf-validation", default=False, action='store_true', help='Execute '
                        'configuration files validation and exit program.')
    parser.add_argument("--keep-locked", "-k", default=False, action='store_true', help='Keep'
                        ' infrastructure locked after task execution.')
    parser.add_argument("--debug", "-d", default=False, action="store_true",
                        help='Print debug output. This option does not affect runner output (use'
                             ' --raw-runner-output for that).')
    parser.add_argument("--syslog", default=False, action="store_true", help='Log warnings and'
                        ' errors to syslog. --debug doesn\'t affect this option!')
    parser.add_argument("--limit", "-l", nargs=1, default=[None], metavar="[LIMIT]",
                        help='Limit task execution to specified host.')
    parser.add_argument("--conf-dir", nargs=1, default=[None], metavar="conf_dir", help='Use '
                        'non-default configuration directory, only allowed for non-binarized exec')
    parser.add_argument("--version", "-v", default=False, action="store_true", help='Display'
                            'app version and exit.')
    parser.add_argument("--raw-runner-output", default=False, action="store_true", help='Print'
                        ' original messages in real time during runner execution.')
    parser.add_argument("--self-setup", nargs=1, default=[None], metavar="LOCAL_SETUP_PATH",
                        help='Setup repo outside of workdir in requested path. This option applies'
                        ' only to infrastructures with allow_user_checkout enabled in infra'
                        ' config!')
    parser.add_argument("--no-color", default=False, action="store_true", help='Disable coloring'
                        'of console messages.')
    parser.add_argument("--no-lock", default=False, action="store_true",
                        help='Do not lock the infrastructure')
    parser.add_argument("--check-mode", "-C", action="store_true", help='Enable --check-mode/-C'
                        ' when using default runner (ansible-playbook).')
    parser.add_argument("--dry-mode", "-D", action="store_true", help='Execute default runner'
                        ' (ansible-playbook) with "ansible_deployer_dry_mode" tag, triggering only'
                        ' required variable validation in pre_tasks. This tag is not predefined!')

    arguments = parser.parse_args(argv)

    if arguments.version:
        version = pkg_resources.require("ansible_deployer")[0].version
        print(f"ansible-deployer version: {version}")
        sys.exit(0)

    if arguments.no_color:
        PRINT_FAIL = PRINT_END = ""
    else:
        PRINT_FAIL = globalvars.PRINT_FAIL
        PRINT_END = globalvars.PRINT_END

    if not arguments.subcommand:
        sub_string = ", ".join(globalvars.SUBCOMMANDS).strip(", ")
        print(f"{PRINT_FAIL}[CRITICAL]: First positional argument (subcommand) is required!"
              f" Available commands are: {sub_string}.{PRINT_END}")
        sys.exit(57)

    options = {}
    options["subcommand"] = arguments.subcommand[0].lower()
    Validators.verify_subcommand(options["subcommand"], arguments.no_color)
    Validators.verify_switches(arguments.subcommand, arguments.no_color)

    options["switches"] = arguments.subcommand[1:]
    options["infra"] = arguments.infrastructure[0]
    options['inventory'] = arguments.inventory[0]
    options["stage"] = arguments.stage[0]
    options["commit"] = arguments.commit[0]
    options["task"] = arguments.task[0]
    options["conf_val"] = arguments.conf_validation
    options["keep_locked"] = arguments.keep_locked
    options["debug"] = arguments.debug
    options["syslog"] = arguments.syslog
    options["limit"] = arguments.limit[0]
    options["raw_output"] = arguments.raw_runner_output
    options["self_setup"] = os.path.abspath(arguments.self_setup[0]) if arguments.self_setup[0] \
                            else None
    options["conf_dir"] = os.path.abspath(arguments.conf_dir[0]) if arguments.conf_dir[0] else None
    options["no_color"] = arguments.no_color
    options["lock"] = not arguments.no_lock
    options["check_mode"] = arguments.check_mode
    options["dry_mode"] = arguments.dry_mode

    return options

def main():
    """ansible-deploy endpoint function"""
    start_ts_raw = datetime.datetime.now()
    start_ts = start_ts_raw.strftime("%Y%m%d_%H%M%S")

    if len(sys.argv) < 2:
        print(f"{globalvars.PRINT_FAIL}[CRITICAL]: Too few arguments{globalvars.PRINT_END}",
              file=sys.stderr)
        sys.exit(2)
    options = parse_options(sys.argv[1:])

    logger = Loggers(options)

    configuration = Config(logger.logger, options["conf_dir"])
    config = configuration.load_configuration()

    if options["subcommand"] in ("run", "verify"):
        workdir = misc.create_workdir(start_ts, configuration.conf, logger.logger)
        log_path = Loggers.set_logging_to_file(logger, workdir, start_ts, configuration.conf)
        logger.flush_memory_handler(True, options["syslog"])
    else:
        logger.flush_memory_handler(False, options["syslog"])

    validators = Validators(logger.logger)
    validators.validate_options(options)
    selected_items = validators.validate_option_values_against_config(config, options)

    if options["subcommand"] in ("run", "verify"):
        os.chdir(options["self_setup"] if options["self_setup"] else workdir)

    user_groups = misc.get_all_user_groups(logger.logger)

    if options["conf_val"]:
        logger.logger.info("Validation of configuration files was successful, program will exit now"
                           ".")
        sys.exit(0)

    if options["subcommand"] == "show":
        misc.show_deployer(config, options)
    else:
        inv_file = misc.get_inventory_file(config, options)
        if not inv_file and options['inventory']:
            inv_file = options['inventory']
        elif options['inventory']:
            logger.logger.info("Ignoring specified inventory file. Using the configured one")
        else:
            logger.logger.fatal("No inventory")

        lockpath = os.path.join(os.path.join(configuration.conf["global_paths"]["work_dir"],
                                "locks") , inv_file.lstrip(f".{os.sep}").replace(os.sep, "_"))
        lock = Locking(logger.logger, options["keep_locked"], (options["infra"], options["stage"]),
                       configuration.conf)
        if options["subcommand"] in ("run", "verify"):
            if not validators.verify_task_permissions(selected_items, user_groups, config):
                logger.logger.critical("Task forbidden")
                sys.exit(errno.EPERM)
            db_connector, db_path = DbSetup.connect_to_db(DbSetup(logger.logger, start_ts,
                                                          configuration.conf, options))
            db_writer = DbWriter(logger.logger, db_connector, db_path)
            runner = Runners(logger.logger, lock, workdir, start_ts_raw,
                             config["tasks"]["setup_hooks"], log_path, db_path)
            if not options["self_setup"]:
                runner.setup_ansible(selected_items["commit"], configuration.conf_dir)
            if options["lock"]:
                lock.lock_inventory(lockpath)
            sequence_record_dict = runner.run_playitem(config, options, inv_file, lockpath,
                                                       db_writer)
            if options["lock"]:
                lock.unlock_inventory(lockpath)
            db_writer.finalize_db_write(sequence_record_dict, False)
        elif options["subcommand"] == "lock":
            lock.lock_inventory(lockpath)
        elif options["subcommand"] == "unlock":
            lock.unlock_inventory(lockpath)

    sys.exit(0)

if __name__ == "__main__":
    main()



"""Setuptools configuration for ansible-deployer"""
from pathlib import Path
from shutil import copytree, copy
from sys import platform, exit
from os import mkdir, environ, walk, sep
from os.path import join, split, exists
from setuptools import setup


CFG_SOURCE = join(Path(__file__).parent, "etc")

def get_cfg_location():
    """Get destination of sample supplementary and config files"""
    if platform == "linux":
        location = join(environ["HOME"], ".ansible-deployer")
    else:
        print("[CRITICAL]: Your platform is not supported!")
        exit(10)

    return location

def create_cfg_location(location: str):
    """Create destination of sample supplementary and config files"""
    try:
        mkdir(location)
    except FileExistsError:
        print(f"[WARNING]: Sample config location already exists at {location}. Copying missing"
              f" files.")
    except Exception as exc:
        print(f"[CRITICAL]: Unable to create sample config location {location} due to {exc}.")
        exit(11)

    return location

def install_configs(src_path: str, dst_path: str):
    """Copy sample supplementary and config files to destination"""
    try:
        copytree(src_path, dst_path)
    except FileExistsError:
        for root, _, files in walk(src_path):
            for file in files:
                full_path = join(root, file)
                if split(full_path)[0] == src_path:
                    file_path = join(dst_path, file)
                    if exists(file_path):
                        print(f"[WARNING]: Sample config file {file} already exists at {dst_path}.")
                    else:
                        copy(full_path, dst_path)
                else:
                    sub_path = join(dst_path, root.replace(src_path, "").strip(sep))
                    file_path = join(sub_path, file)
                    try:
                        mkdir(sub_path)
                        if exists(file_path):
                            print(f"[WARNING]: Sample Config file {file} already exists at"
                                  f" {sub_path}.")
                        else:
                            copy(full_path, sub_path)
                    except FileExistsError:
                        print(f"[WARNING]: Sample config dir {sub_path} already exists.")
                    except Exception as exc:
                        print(f"[WARNING]: Unable to install sample config file {file} due to"
                              f" {exc}.")
                        exit(13)
    except Exception as exc:
        print(f"[CRITICAL]: Unable to install sample config files due to {exc}.")
        exit(12)


if __name__ == "__main__":
    location = get_cfg_location()
    create_cfg_location(location)
    install_configs(CFG_SOURCE, location)
    setup()



"""Miscellaneous helper functions"""

import os
import sys
import grp


def create_parent_workdir(short_ts: str, date_dir: str, conf: dict, logger):
    """Function to create parent working directory"""
    if short_ts not in os.listdir(conf["global_paths"]["work_dir"]):
        try:
            os.mkdir(date_dir)
            try:
                os.chmod(date_dir, int(conf["permissions"]["parent_workdir"], 8))
                logger.debug("Successfully created parent work dir: %s", date_dir)
            except Exception as exc:
                logger.critical("Failed to change permissions of parent work dir: %s error was: %s",
                                date_dir, exc)
                sys.exit(90)
        except Exception as exc:
            logger.critical("Failed to create parent work dir: %s error was: %s", date_dir, exc)
            sys.exit(90)

def create_workdir(timestamp: str, conf: dict, logger):
    """
    Function to create working directory on file system, we expect it to change
    the cwd to newly created workdir at the end.
    """
    short_ts = timestamp.split("_")[0]
    date_dir = os.path.join(conf["global_paths"]["work_dir"], short_ts)
    base_dir = os.path.join(date_dir, conf["global_paths"]["sequences_subdir"])
    create_parent_workdir(short_ts, date_dir, conf, logger)

    #
    #TODO: Add locking of the directory
    if conf["global_paths"]["sequences_subdir"] not in os.listdir(date_dir):
        seq_path = os.path.join(base_dir, f"{conf['file_naming']['sequence_prefix']}0000")
        try:
            os.mkdir(base_dir)
            try:
                os.chmod(base_dir, int(conf["permissions"]["parent_workdir"], 8))
                logger.debug("Successfully created parent work dir: %s", base_dir)
            except Exception as exc:
                logger.critical("Failed to change permissions of parent work dir: %s error was: %s",
                                base_dir, exc)
                sys.exit(90)
        except Exception as exc:
            logger.critical("Failed to create parent work dir: %s error was: %s", base_dir, exc)
            sys.exit(90)
    else:
        sequence_list = os.listdir(base_dir)
        sequence_list.sort()
        new_sequence = int(sequence_list[-1].split(conf['file_naming']['sequence_prefix'])[1]) + 1
        seq_path = os.path.join(base_dir, f"{conf['file_naming']['sequence_prefix']}"
                                          f"{new_sequence:04d}")

    try:
        os.mkdir(seq_path)
        os.chmod(seq_path, int(conf["permissions"]["workdir"], 8))
    except Exception as e:
        logger.critical("Failed to create work dir: %s error was: %s", seq_path, e, file=sys.stderr)
        sys.exit(91)
    logger.debug("Successfully created workdir: %s", seq_path)
    return seq_path

# TODO: At least infra level should be returned from validate options since we do similar check
# (existence) there.
def get_inventory_file(config: dict, options: dict):
    """
    Function returning relativ path to inventory file.
    :param config:
    :param options:
    :return:
    """

    inv_file = None

    for item in config["infra"]:
        if item["name"] == options["infra"]:
            for elem in item["stages"]:
                if elem["name"] == options["stage"]:
                    inv_file = elem["inventory"]

    return inv_file

def get_all_user_groups(logger):
    """
    Function returning all user groups in form of their names
    """
    user_groups = [grp.getgrgid(g).gr_name for g in os.getgroups()]
    logger.debug("User groups:%s %s", user_groups, grp.getgrgid(os.getgid()).gr_name)
    user_groups.append(str(grp.getgrgid(os.getgid()).gr_name))


    return user_groups

def show_deployer(config: dict, options: dict):
    """
    Function showing available configs: tasks, infrastructures, permissions.
    """
    content = {}

    if not options["switches"] or "all" in options["switches"]:
        options["switches"] = ["task", "infra"]

    if "task" in options["switches"]:
        content["tasks"] = []
        for item in config["tasks"]["tasks"]:
            content["tasks"].append(item["name"])

    if "infra" in options["switches"]:
        content["infrastructures"] = []
        for item in config["infra"]:
            content["infrastructures"].append(item["name"])

    print(format_show_deployer(content))

def format_show_deployer(print_data: dict):
    """
    Format data from show_deployer function to output
    """
    content = ""
    for key, value in print_data.items():
        values = ", ".join(value).strip(", ")
        content = f"{content}\n\nAvailable {key}:\n{values}\n"

    return content
